-- new sdram controller now has dm pin available to truncate burst write
-- initialized to fullpage burst read/write.
-- actual burst is limited to 64 32-bit words
-- memory is divided into two regions:
-- a) HTR event buffer
--   each event has 16 512-word HTR buffers
--   there are 128 such buffers
-- absolute address range: 0x??0000 - 0x??7fff in bytes
-- b) monitoring event buffer
--   each event is 8k-word long, a total of 128 buffers are available
-- to access the monitoring data, a PCI memory space of 32k-bytes
-- is assigned. Every event readout starts at relative address of 0x0
-- to fully use the memory available, memory address are calculated
-- with multipliers and adders.
-- absolute address range 0x??8000-0x??ffff in bytes
-- Except for PCI3 accesses, all read/write starts at addresses aligned to
-- 64 word boundary ( six LSBs all zero )
library IEEE,mylib;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use mylib.vtx_pack.all;
use mylib.dcc_pack.all;
library UNISIM;
use UNISIM.VComponents.all;

entity ddr_w_queue is
    port (
-- reset input, should be connected to the inverted LOCK output of dll
        reset: in STD_LOGIC;
-- system clock input
        clk: in STD_LOGIC;
-- fill memory with random numbers
        mem_test: in STD_LOGIC;
-- htr and monitored event write address
		  w_addr : IN std_logic_vector(11 downto 0);
-- monitored event data
-- LRB data read in
-- lrbx_avl signals 64 data words available or data includes trailer
        event_avl: in STD_LOGIC;
		  lrb1_avl : IN std_logic;
		  lrb2_avl : IN std_logic;
-- lrbx_we signals lrbx_data are valid
        event_we: in STD_LOGIC;
		  lrb1_we : IN std_logic;
		  lrb2_we : IN std_logic;
-- lrbx_re starts the burst read from corresponding LRB buffer
        event_re: out STD_LOGIC;
		  lrb1_re : OUT std_logic;
		  lrb2_re : OUT std_logic;
		  src_sel : OUT std_logic_vector(1 downto 0);
-- bit 33 and 32 used to indicate header and trailer of the block
		  lrb_data : IN std_logic_vector(33 downto 0);
        event_data: in STD_LOGIC_VECTOR (33 downto 0);
-- queue signals
		  ddr_w_gnt: in std_logic;
		  ddr_w_dav: out std_logic;
		  ddr_w_req: out std_logic;
        ddr_w_length: out STD_LOGIC_VECTOR (5 downto 0);
        ddr_w_addr: out STD_LOGIC_VECTOR (20 downto 0);
        ddr_w_data: out STD_LOGIC_VECTOR (31 downto 0);
-- download data write
        w_data: in STD_LOGIC_VECTOR (31 downto 0);
        rd_debug: in STD_LOGIC;
        r_data: out STD_LOGIC_VECTOR (31 downto 0);
-- download address
        addr: in STD_LOGIC_VECTOR (13 downto 0);
        run: in STD_LOGIC;
        is_high: in STD_LOGIC;
        n_strobe: in STD_LOGIC;
        pci3_cycle: in STD_LOGIC;
        wr_ddr_data: in STD_LOGIC;
		  ddr_pa: in std_logic_vector(6 downto 0)
    );
end ddr_w_queue;

architecture my_arch of ddr_w_queue is
signal strobe: std_logic := '0';
signal idle: std_logic := '1';
-- write mux signals
signal lrb_evt_we: std_logic := '0';
signal any_we: std_logic := '0';
signal src_sel_i: std_logic_vector(1 downto 0) := (others => '0');
signal mon_vme_data: std_logic_vector(33 downto 0) := (others => '0');
signal wfifo_din: std_logic_vector(33 downto 0) := (others => '0');
-- address file
signal first_vme: std_logic := '1';
signal first_write: std_logic := '0';
signal w_data_valid: std_logic := '0';
signal vme_addr: std_logic := '0';
signal vme_pending: std_logic := '0';
signal wr_waddr: std_logic := '0';
signal addr_0: std_logic := '0';
signal waddr_in: std_logic_vector(20 downto 0) := (others => '0');
signal wr_length: std_logic := '0';
signal length_in: std_logic_vector(5 downto 0) := (others => '0');
signal rst_length: std_logic := '0';
signal empty_length: std_logic := '0';
signal full_length: std_logic := '0';
signal wr_mid_addr: std_logic := '0';
signal new_mid: std_logic := '0';
signal mid_addr: std_logic_vector(7 downto 0) := (others => '0');
signal mid_addr_in: std_logic_vector(7 downto 0) := (others => '0');
signal mid_s: std_logic_vector(7 downto 0) := (others => '0');
signal mid_ci: std_logic_vector(8 downto 0) := (others => '0');
signal mid_addr_a: std_logic_vector(10 downto 0) := (others => '0');
-- random number memory test signals
signal lfsr : std_logic_vector(31 downto 0) := x"00000000";
signal tst_addr : std_logic_vector(7 downto 0) := x"00";
signal ec_lfsr: std_logic := '0';
-------------
signal debug: std_logic_vector(31 downto 0) := (others => '0');
begin
ddr_w_req <= not empty_length;
src_sel <= src_sel_i;
process(clk)
begin
  if(clk'event and clk = '1')then
    strobe <= n_strobe;
	 if(strobe = '1')then
	   w_data_valid <= wr_ddr_data and not w_data_valid;
	 end if;
  end if;
end process;
-- write queue priority encoder
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
	   src_sel_i <= "00";
	 elsif(idle = '1')then
		src_sel_i(1) <= run and (lrb1_avl or lrb2_avl) and not event_avl;
		src_sel_i(0) <= run and (lrb2_avl or event_avl);
    end if;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
	   idle <= '1';
	   lrb1_re <= '0';
	   lrb2_re <= '0';
	   event_re <= '0';
-- ddr fifo write enable
      lrb_evt_we <= '0';
      any_we <= '0';
      addr_0 <= '0';
      ddr_w_dav <= '0';
      wr_length <= '0';
      vme_pending <= '0';
      rst_length <= '0';
	 else
		idle <= (idle and not(lrb1_avl or lrb2_avl or event_avl)) or wr_length or full_length;
	   lrb1_re <= idle and lrb1_avl and not event_avl and not lrb2_avl and not full_length;
	   lrb2_re <= idle and lrb2_avl and not event_avl and not full_length;
	   event_re <= idle and event_avl and not full_length;
-- ddr fifo write enable
      lrb_evt_we <= (event_we or lrb1_we or lrb2_we) and run;
      any_we <= lrb_evt_we or (w_data_valid and strobe);
-- if vme starting write address is odd, add one more dummy word before the valid data word
      if(run = '1' or mem_test = '1')then
		  addr_0 <= '0';
      elsif(wr_waddr = '1')then
		  addr_0 <= addr(0);
		end if;
      ddr_w_dav <= lrb_evt_we or (w_data_valid and strobe)	or
		   (wr_length and not length_in(0) and not addr_0) or (not mem_test and vme_addr and addr(0));
		if(run =	'1')then
        wr_length <= any_we and not lrb_evt_we;
		elsif(mem_test = '0')then
        wr_length <= vme_pending and not pci3_cycle and not wr_ddr_data and strobe;
		else
        wr_length <= vme_pending and not wr_ddr_data and strobe;
		end if;
		vme_pending <= vme_addr or (vme_pending and not wr_length); 
-- rst_length coincides with the first write
-- If starting address is odd, full page bursts backward. Must make it a two word
-- burst (with a dummy word padded before the valid data word)
      if(run = '1')then
		  rst_length <= not any_we and lrb_evt_we;
		else
		  rst_length <= wr_waddr;
		end if;
    end if;
  end if;
end process;
-- write queue data path
-- src_sel_i = "1x" lrb_data
-- src_sel_i = "01" event_data
-- src_sel_i = "00" vme_wdata
process(clk)
begin
  if(clk'event and clk = '1')then
    if(src_sel_i(0) = '1')then
      mon_vme_data <= event_data;
    elsif(mem_test = '0')then
      mon_vme_data <= "00" & w_data;
	 else
      mon_vme_data <= "00" & lfsr;
    end if;
    if(src_sel_i(1) = '1')then
      wfifo_din(32 downto 0) <= lrb_data(32 downto 0);
	 else
      wfifo_din(32 downto 0) <= mon_vme_data(32 downto 0);
    end if;
-- make sure never overwrite initial values
    if(first_write = '1')then
      wfifo_din(33) <= '0';
    elsif(src_sel_i(1) = '1')then
      wfifo_din(33) <= lrb_data(33);
	 else
      wfifo_din(33) <= mon_vme_data(33);
    end if;
  end if;
end process;
ddr_w_data <= wfifo_din(31 downto 0);
i_mid_addr: ramb16_s9
-- this block ram needs initialization	for INIT_02 and INIT_03
     generic map(
	  		INIT_02 => x"0000000000000000000000000000008000000000000000000000000000000000",
	      INIT_03 => x"0000000000000000000000000000000078706860585048403830282018100800")
     port map(
         di => mid_addr_in,
         dip => "0",
         addr => mid_addr_a,
         clk => clk,
         ssr => '0',
         en => '1',
         we => wr_mid_addr,
         do => mid_addr);
-- header
new_mid <= wfifo_din(33);
-- mid_addr is address in 64-words block. new_mid when headers seen
mid_addr_a(10 downto 5) <= "0000" & new_mid & src_sel_i(1);
mid_addr_a(4 downto 0) <= src_sel_i(0) & "0000" when src_sel_i(1) = '0' else w_addr(4 downto 0);
i_mid_addr_in: add_oneN 
    generic map( N => 8)
    port map(d => mid_addr, rst => '0', sum => mid_addr_in);
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
      wr_mid_addr <= '0';
      first_write <= '0';
      first_vme <= '1';
      vme_addr <= '0';
      wr_waddr <= '0';
	 else
      wr_mid_addr <= first_write;
      first_write <= not any_we and lrb_evt_we;
		first_vme <= (first_vme and not vme_addr) or wr_length;
		vme_addr <= first_vme and wr_ddr_data and strobe;
	   wr_waddr <= (first_write and run) or (not run and vme_addr);
	 end if;
  end if;
end process;
waddr_in <= w_addr(11 downto 5) & mid_addr & "000000" when run = '1'
            else ddr_pa & addr when mem_test = '0'
				else ddr_pa & tst_addr & "000000";
i_waddr : srl_fiforeg
generic map (N => 21)
  port map(
	din => waddr_in,
	write => wr_waddr,
	read => ddr_w_gnt, 
	clk => clk,
	rst => reset,
	dout => ddr_w_addr);
process(clk)
begin
  if(clk'event and clk = '1')then
-- length is actual length minus one, reset at the first write
    if(rst_length = '1')then
	   length_in <= "000000";
	 elsif(any_we = '1')then
		length_in <= length_in + 1;
    end if;
  end if;
end process;
i_length : srl_fiforeg
generic map (N => 6)
  port map(
	din => length_in,
	write => wr_length,
	read => ddr_w_gnt, 
	clk => clk,
	rst => reset,
	dout => ddr_w_length,
-- wr_length always comes after wr_waddr, so use it as w_req
	empty => empty_length,
	full => full_length);
-- random number memory test added 8/3/2005
-- random number generator
process(clk)
begin
  if(clk'event and clk = '1')then
    if(mem_test = '0')then
	   tst_addr <= (others => '0');
	 elsif(wr_waddr = '1')then
	   tst_addr <= tst_addr + 1;
	 end if;
    if(mem_test = '0')then
	   lfsr <= (others => '0');
	 elsif(ec_lfsr = '1')then
	   lfsr <= lfsr(30 downto 0) & not(lfsr(31) xor lfsr(21) xor lfsr(1) xor lfsr(0));
	 end if;
	 ec_lfsr <= mem_test and w_data_valid and strobe;
  end if;
end process;
-- debug
process(clk)
begin
  if(clk'event and clk = '1')then
    if(strobe = '1')then
		debug(0) <= event_avl;
		debug(1) <= lrb1_avl;
		debug(2) <= lrb2_avl;
		debug(3) <= event_we;
		debug(4) <= lrb1_we;
		debug(5) <= lrb2_we;
		debug(6) <= idle;
		debug(7) <= full_length;
		debug(8) <= empty_length;
		debug(10 downto 9) <= src_sel_i;
    end if;
  end if;
end process;
debug(15 downto 11) <= "00000";
r_data(31 downto 16) <= "ZZZZZZZZZZZZZZZZ";
r_data(15 downto 0) <= debug(15 downto 0) when rd_debug = '1' and addr(2 downto 0) = "011" else "ZZZZZZZZZZZZZZZZ";
end my_arch;
